#!/usr/bin/env python3

#
# Author: Bioinformatics group, Takara Bio USA, Inc.
#
# Script to generate summary files and report from STAR fusion outputs
# This script receives the output folder from STAR fusion as input
#
# -----------------------------------------------------------------------------
import argparse
import os
from collections import defaultdict
import sys
import pathlib
import re
import pandas as pd
import glob
import scipy.sparse
from scipy.io import mmwrite


DESCRIPTION = """Generate summary matrix files and reports from STAR fusion outputs."""

###
# Script information
###

# script info
SCRIPT_NAME = "fusion_report.py"
VERSION = "3.1.22"


###
# Parse and define arguments
###


def _parse_command_line_args(sys_argv):
    """Takes in command line parameters and parses them for use

    Args:
        sys_argv: command line input

    Returns:
        args (list) : input argument values
    """
    parser = argparse.ArgumentParser(
        description=DESCRIPTION, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    required = parser.add_argument_group("required arguments")

    required.add_argument(
        "-i",
        "--input_dir",
        dest="input_dir",
        help="Full path to directory containing STAR Fusion outputs.",
        required=True,
    )

    required.add_argument(
        "-o",
        "--output_dir",
        dest="output_dir",
        help="Full path to output directory to store matrices.",
        required=True,
    )

    args, unknown = parser.parse_known_args(sys_argv)
    return args


def get_barcode(sample_name):
    if re.match(r"(.*)[ACGTN]{16,24}(.*)", sample_name):
        if re.match(r".*_L00\d(_trimmed)?$", sample_name):
            curr_barcode = sample_name.split("_")[-3]
        else:
            curr_barcode = sample_name.split("_")[-2]
    else:
        curr_barcode = sample_name.replace("_trimmed", "")
    return curr_barcode


# ---------- main ---------- #
def main(sys_argv):
    args = _parse_command_line_args(sys_argv)

    # get list of star fusion result folders
    master_dir = args.input_dir
    out_dir = args.output_dir

    # ubfolder_names = os.listdir(master_dir)
    fusion_result_files = os.path.join(master_dir, "*_fusion_predictions.abridged.tsv")
    fusion_results_list = glob.glob(fusion_result_files, recursive=True)

    # get barcode from folder name
    barcode_list = [
        i.split("/")[-1].replace("_fusion_predictions.abridged.tsv", "")
        for i in fusion_results_list
    ]
    final_barcode_list = [get_barcode(x) for x in barcode_list]
    # initialize dictionaries for fusion and gene matrices
    junction_dict = defaultdict(dict)
    spanning_dict = defaultdict(dict)

    # loop through the fusion predictions file for each sample
    for result_file in fusion_results_list:
        sample_name = result_file.split("/")[-1].replace(
            "_fusion_predictions.abridged.tsv", ""
        )
        curr_barcode = get_barcode(sample_name)

        with open(result_file) as result:
            lines = result.readlines()
            for line in lines:
                li = line.strip()
                if not li.startswith("#"):
                    # get junction and spanning counts and save to dictionary
                    fusion_genes = li.split("\t")[0]
                    junction_ct = li.split("\t")[1]
                    spanning_ct = li.split("\t")[2]
                    junction_dict[fusion_genes][curr_barcode] = int(junction_ct)
                    spanning_dict[fusion_genes][curr_barcode] = int(spanning_ct)


    pathlib.Path(os.path.join(args.input_dir, out_dir, "junction")).mkdir(
        parents=True, exist_ok=True
    )
    
    pathlib.Path(os.path.join(args.input_dir, out_dir, "span")).mkdir(
        parents=True, exist_ok=True
    )


    # Define the output file path for the .mtx file
    junction_matrix_file = os.path.join(out_dir, "junction/matrix.mtx")
    junction_features_file = os.path.join(out_dir, "junction/features.tsv")
    junction_barcodes_file = os.path.join(out_dir, "junction/barcodes.tsv")
    spanning_matrix_file = os.path.join(out_dir, "span/matrix.mtx")
    spanning_features_file = os.path.join(out_dir, "span/features.tsv")
    spanning_barcodes_file = os.path.join(out_dir, "span/barcodes.tsv")

    # Construct the matrix data
    rows = []
    cols = []
    jxn_data = []

    # Iterate over genes and barcodes to create the sparse matrix (CSR format)
    for gene_index, gene in enumerate(junction_dict):
        for barcode_index, barcode in enumerate(final_barcode_list):
            value = junction_dict.get(gene, {}).get(barcode, 0.0)
            if value != 0.0:  # Only store non-zero values for efficiency
                rows.append(gene_index)
                cols.append(barcode_index)
                jxn_data.append(value)

    # Create the sparse matrix in CSR format
    num_genes = len(junction_dict)
    num_barcodes = len(final_barcode_list)
    junction_matrix_sparse = scipy.sparse.csr_matrix((jxn_data, (rows, cols)), shape=(num_genes, num_barcodes))

    # Write the matrix to the .mtx file
    mmwrite(junction_matrix_file, junction_matrix_sparse)

    # Add gene and barcode names as headers (these will be stored separately in a separate file, as Matrix Market format doesn't support attributes)
    with open(junction_features_file, 'w') as gene_file:
        for gene in junction_dict.keys():
            gene_file.write(gene + '\t')
            gene_file.write(gene + '\n')

    with open(junction_barcodes_file, 'w') as barcode_file:
        for barcode in final_barcode_list:
            barcode_file.write(barcode + '\n')

    # Construct the matrix data
    rows = []
    cols = []
    span_data = []

    # Iterate over genes and barcodes to create the sparse matrix (CSR format)
    for gene_index, gene in enumerate(spanning_dict):
        for barcode_index, barcode in enumerate(final_barcode_list):
            value = spanning_dict.get(gene, {}).get(barcode, 0.0)
            if value != 0.0:  # Only store non-zero values for efficiency
                rows.append(gene_index)
                cols.append(barcode_index)
                span_data.append(value)

    # Create the sparse matrix in CSR format
    num_genes = len(spanning_dict)
    num_barcodes = len(final_barcode_list)
    spanning_matrix_sparse = scipy.sparse.csr_matrix((span_data, (rows, cols)), shape=(num_genes, num_barcodes))

    # Write the matrix to the .mtx file
    mmwrite(spanning_matrix_file, spanning_matrix_sparse)

    # Add gene and barcode names as headers (these will be stored separately in a separate file, as Matrix Market format doesn't support attributes)
    with open(spanning_features_file, 'w') as gene_file:
        for gene in spanning_dict.keys():
            gene_file.write(gene + '\t')
            gene_file.write(gene + '\n')

    with open(spanning_barcodes_file, 'w') as barcode_file:
        for barcode in final_barcode_list:
            barcode_file.write(barcode + '\n')


if __name__ == "__main__":
    main(sys.argv[1:])
