#!/usr/bin/env python3

#
# Author: Bioinformatics group, Takara Bio USA, Inc.
#
# Script to generate summary files and report from Cogent Analyze outputs
#
# -----------------------------------------------------------------------------
import argparse
import os
import re
from collections import defaultdict
import sys
import glob
import json
import pandas as pd
from pathlib import Path

DESCRIPTION = (
    """Generate summary matrix files and reports from Cogent Analyze outputs."""
)

###
# Script information
###

# script info
SCRIPT_NAME = "create_analyze_stats.py"
VERSION = "3.1.22"

###
# Parse and define arguments
###


def _parse_command_line_args(sys_argv):
    """Takes in command line parameters and parses them for use

    Args:
        sys_argv: command line input

    Returns:
        args (list) : input argument values
    """
    parser = argparse.ArgumentParser(
        description=DESCRIPTION, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    required = parser.add_argument_group("required arguments")

    required.add_argument(
        "-i",
        "--input_dir",
        dest="input_dir",
        help="Full path to directory containing all the relevant logs.",
        required=True,
    )

    required.add_argument(
        "-t",
        "--type_of_experiment",
        dest="type_of_exp",
        help="Experiment type",
        required=True,
    )

    args, unknown = parser.parse_known_args(sys_argv)
    return args


def get_barcode(sample_name, index_1, index_2):
    if re.match(r"(.*)[ACGTN]{16,24}(.*)", sample_name):
        if re.match(r".*_L00\d(_trimmed)?$", sample_name):
            curr_barcode = sample_name.split("_")[index_1]
        else:
            curr_barcode = sample_name.split("_")[index_2]
    else:
        curr_barcode = sample_name.replace("_trimmed", "")
    return curr_barcode


# ---------- main ---------- #
def main(sys_argv):
    args = _parse_command_line_args(sys_argv)

    master_dir = args.input_dir
    exp_type = args.type_of_exp

    stats_dict = defaultdict(dict)

    # parse through cutadapt logs
    cutadapt_log_files = os.path.join(master_dir, "*.cutadapt.json")

    cutadapt_log_list = glob.glob(cutadapt_log_files, recursive=True)

    for cutadapt_log in cutadapt_log_list:

        sample_name = cutadapt_log.split("/")[-1].replace(".cutadapt.json", "")
        if re.match(r"(.*)[ACGTN]{16,24}(.*)", sample_name):
            if re.match(r".*_L00\d(_trimmed)?$", sample_name):
                curr_barcode = sample_name.split("_")[-2]
            else:
                curr_barcode = sample_name.split("_")[-1]

            sample = sample_name.replace("_" + curr_barcode, "")
            stats_dict[curr_barcode]["Barcode"] = curr_barcode
        else:
            curr_barcode = sample_name
            sample = sample_name
            stats_dict[curr_barcode]["Barcode"] = "NA"

        print("Parsing cutadapt logs for ", curr_barcode)
        with open(cutadapt_log) as c_log:

            data = json.load(c_log)

            stats_dict[curr_barcode]["Sample"] = sample
            stats_dict[curr_barcode]["Barcoded_Reads"] = data["read_counts"]["input"]
            stats_dict[curr_barcode]["Trimmed_Reads"] = data["read_counts"]["output"]

    # parse through star logs
    star_log_files = os.path.join(master_dir, "*_Log.final.out")

    if glob.glob(star_log_files, recursive=True):
        star_log_list = glob.glob(star_log_files, recursive=True)

        for star_log in star_log_list:

            sample_name = star_log.split("/")[-1].replace("_Log.final.out", "")
            curr_barcode = get_barcode(sample_name, -3, -2)
            print("Parsing STAR logs for ", curr_barcode)
            unmapped_count = 0
            multimapped_count = 0
            with open(star_log) as log_file:
                for line in log_file:
                    if "Uniquely mapped reads number" in line:
                        line = line.strip()
                        stats_dict[curr_barcode]["Uniquely_Mapped_Reads"] = int(
                            line.split("\t")[-1]
                        )
                    if "Number of reads mapped to multiple loci" in line:
                        line = line.strip()
                        multimapped_count += int(line.split("\t")[-1])
                    if "Number of reads unmapped:" in line:
                        line = line.strip()
                        unmapped_count += int(line.split("\t")[-1])
                    if "Number of reads mapped to too many loci" in line:
                        line = line.strip()
                        multimapped_count += int(line.split("\t")[-1])
                    if "Number of chimeric reads" in line:
                        line = line.strip()
                        stats_dict[curr_barcode]["Chimeric_Reads"] = int(
                            line.split("\t")[-1]
                        )

            stats_dict[curr_barcode]["Unmapped_Reads"] = unmapped_count
            stats_dict[curr_barcode]["Multimapped_Reads"] = multimapped_count
            stats_dict[curr_barcode]["Mapped_Reads"] = (
                stats_dict[curr_barcode]["Uniquely_Mapped_Reads"]
                + stats_dict[curr_barcode]["Multimapped_Reads"]
            )
    # parse through reads per gene files from star
    rev_exp_list = [
        "stranded",
        "stranded_umi",
        "shasta_total_rna",
    ]

    unstranded_exp_list = [
        "smartseq_fla",
        "smartseq_fla_umi",
        "icell8_fla",
        "smartseq_pro",
    ]

    if exp_type in rev_exp_list:
        reads_per_gene = os.path.join(master_dir, "*ReadsPerGene.out.tab")

        if glob.glob(reads_per_gene, recursive=True):
            reads_file_list = glob.glob(reads_per_gene, recursive=True)

            for reads_file in reads_file_list:

                sample_name = reads_file.split("/")[-1].replace(
                    "_ReadsPerGene.out.tab", ""
                )
                curr_barcode = get_barcode(sample_name, -3, -2)
                print("Parsing STAR Reads Per Gene file for ", curr_barcode)
                fwd_count = 0
                rev_count = 0
                unstranded_count = 0

                with open(reads_file) as reads_log:
                    for _ in range(4):
                        next(reads_log)
                    for line in reads_log:
                        # Strip whitespace and skip empty lines
                        line = line.strip()
                        if not line:
                            continue  # Skip empty lines
                        try:
                            (id, unstranded, fwd, rev) = line.split("\t")
                            unstranded_count += int(unstranded)
                            fwd_count += int(fwd)
                            rev_count += int(rev)
                        except ValueError:
                            continue      
                
                try:
                    # take average of the counts - see here: https://groups.google.com/g/rna-star/c/gZRJx3ElRNo
                    stats_dict[curr_barcode]["Strand_Specificity"] = round(
                        rev_count / ((rev_count + fwd_count + unstranded_count) / 2), 3
                    )
                
                    if stats_dict[curr_barcode]["Strand_Specificity"] > 1.0:
                        stats_dict[curr_barcode]["Strand_Specificity"] = 1.0
                except Exception as e:
                    stats_dict[curr_barcode]["Strand_Specificity"] = 'NA'

                header = [
                    "Barcode",
                    "Sample",
                    "Barcoded_Reads",
                    "Trimmed_Reads",
                    "Unmapped_Reads",
                    "Mapped_Reads",
                    "Uniquely_Mapped_Reads",
                    "Multimapped_Reads",
                    "Chimeric_Reads",
                    "Exon_Reads",
                    "Intron_Reads",
                    "Gene_Reads",
                    "Intergenic_Reads",
                    "No_of_Genes",
                    "No_of_Transcripts",
                    "Ribosomal_Reads",
                    "Mitochondrial_Reads",
                    "Strand_Specificity",
                ]

                if exp_type == "stranded_umi":
                    header = [
                        "Barcode",
                        "Sample",
                        "Barcoded_Reads",
                        "Trimmed_Reads",
                        "Unmapped_Reads",
                        "Mapped_Reads",
                        "Uniquely_Mapped_Reads",
                        "Multimapped_Reads",
                        "Chimeric_Reads",
                        "No_of_Reads_w_UMIs",
                        "Reads_After_Dedup_nUMIs_USSs",
                        "Exon_nUMIs_USSs",
                        "Intron_nUMIs_USSs",
                        "Gene_nUMIs_USSs",
                        "Intergenic_Reads",
                        "No_of_Genes_nUMIs_USSs",
                        "No_of_Transcripts_nUMIs_USSs",
                        "Ribosomal_Reads",
                        "Mitochondrial_Reads",
                        "Strand_Specificity",
                    ]

    elif exp_type in unstranded_exp_list:
        header = [
            "Barcode",
            "Sample",
            "Barcoded_Reads",
            "Trimmed_Reads",
            "Unmapped_Reads",
            "Mapped_Reads",
            "Uniquely_Mapped_Reads",
            "Multimapped_Reads",
            "Chimeric_Reads",
            "Exon_Reads",
            "Intron_Reads",
            "Gene_Reads",
            "Intergenic_Reads",
            "No_of_Genes",
            "No_of_Transcripts",
            "Ribosomal_Reads",
            "Mitochondrial_Reads",
        ]

    # parse through sortmerna logs
    sortmerna_log_files = os.path.join(master_dir, "*.sortmerna.log")

    if glob.glob(sortmerna_log_files, recursive=True):
        sortmerna_log_list = glob.glob(sortmerna_log_files, recursive=True)

        for sortmerna_log in sortmerna_log_list:

            sample_name = sortmerna_log.split("/")[-1].replace(".sortmerna.log", "")
            curr_barcode = get_barcode(sample_name, -3, -2)
            print("Parsing sortmerna logs for ", curr_barcode)
            with open(sortmerna_log) as s_log:
                for line in s_log:
                    if "Total reads passing E-value" in line:
                        stats_dict[curr_barcode]["Ribosomal_Reads"] = int(
                            int(line.split(" ")[-2]) / 2
                        )

    else:
        for barcode in stats_dict:
            stats_dict[barcode]["Ribosomal_Reads"] = "NA"

    # parse through barcode stats file
    try:
        barcode_stats_file = glob.glob(
            os.path.join(master_dir, "barcode_stats.csv")
        )
        barcode_stats = pd.read_csv(barcode_stats_file[0], sep=",", index_col = 0)
        barcodes_list = barcode_stats.index.tolist()
        for barcode in barcodes_list:
            print("Parsing barcode stats for ", barcode)
            stats_dict[barcode]["Exon_Reads"] = round(barcode_stats['Exon_Only_Counts'][barcode])
            stats_dict[barcode]["Intron_Reads"] = round(barcode_stats['Intron_Only_Counts'][barcode])
            stats_dict[barcode]["Gene_Reads"] = round(barcode_stats['Exon_plus_Intron_Counts'][barcode])
            stats_dict[barcode]["Intergenic_Reads"] = (
                stats_dict[barcode]["Mapped_Reads"] - stats_dict[barcode]["Gene_Reads"]
            )
            stats_dict[barcode]["Mitochondrial_Reads"] = round(barcode_stats['Mitochondrial_Reads'][barcode])
            stats_dict[barcode]["No_of_Genes"] = round(barcode_stats['Number_of_Genes_Expressed'][barcode])
            stats_dict[barcode]["No_of_Transcripts"] = round(barcode_stats['Number_of_Isoforms_Expressed'][barcode])
            
        del barcode_stats

    except IOError:
        print("Barcode stats file does not exist.")
        sys.exit(1)

    # parse umitools log
    if exp_type == "smartseq_fla_umi" or exp_type == "stranded_umi":
        umitools_log_files = os.path.join(master_dir, "*umitools.log")
        read_count = ""
        if glob.glob(umitools_log_files, recursive=True):

            umitools_log_list = glob.glob(umitools_log_files, recursive=True)
            for umitools_log in umitools_log_list:

                sample_name = umitools_log.split("/")[-1].replace("_umitools.log", "")
                curr_barcode = get_barcode(sample_name, -3, -2)
                print("Parsing umitools logs for ", curr_barcode)
                with open(umitools_log) as u_log:
                    for line in u_log:
                        if "Number of reads out:" in line:
                            stats_dict[curr_barcode]["Reads_After_Dedup_nUMIs_USSs"] = (
                                int(line.split(" ")[-1])
                            )
                        if "INFO Reads: Input Reads:" in line:
                            if exp_type == "stranded_umi":
                                read_count = line.split(" ")[6]
                                stats_dict[curr_barcode]["No_of_Reads_w_UMIs"] = int(
                                    read_count.replace(",", "")
                                )
                            else:
                                read_count = line.split(" ")[-1]
                                stats_dict[curr_barcode]["No_of_Reads_w_UMIs"] = int(
                                    read_count
                                )
                if exp_type == "stranded_umi":
                    stats_dict[curr_barcode]["Intergenic_Reads"] = (
                        stats_dict[curr_barcode]["Reads_After_Dedup_nUMIs_USSs"] - stats_dict[curr_barcode]["Gene_Reads"]
                    )

    # special case for smartseq_fla_umi kits, needs additional stats file with umis alone
    if exp_type == "smartseq_fla_umi":
        # parse through umi barcode stats
        try:
            barcode_stats_umi_file = glob.glob(
                os.path.join(master_dir, "barcode_stats_umis.csv")
            )
            barcode_stats_umi = pd.read_csv(barcode_stats_umi_file[0], sep=",", index_col = 0)

            barcodes_list_umi = barcode_stats_umi.index.tolist()
            for barcode in barcodes_list_umi:
                print("Calculating umitools barcode stats for ", barcode)
                stats_dict[barcode]["Exon_nUMIs_USSs"] = round(barcode_stats_umi['Exon_Only_Counts'][barcode])
                stats_dict[barcode]["Intron_nUMIs_USSs"] = round(barcode_stats_umi['Intron_Only_Counts'][barcode])
                stats_dict[barcode]["Gene_nUMIs_USSs"] = round(barcode_stats_umi['Exon_plus_Intron_Counts'][barcode])

                stats_dict[barcode]["No_of_Genes_nUMIs_USSs"] = round(barcode_stats_umi['Number_of_Genes_Expressed'][barcode])
                stats_dict[barcode]["No_of_Transcripts_nUMIs_USSs"] = round(barcode_stats_umi['Number_of_Isoforms_Expressed'][barcode])
                stats_dict[barcode]["Intergenic_Reads_nUMI_USSs"] = (
                    stats_dict[curr_barcode]["Reads_After_Dedup_nUMIs_USSs"] - stats_dict[curr_barcode]["Gene_nUMIs_USSs"]
                )

        except IOError:
            print("Barcode UMI stats file does not exist.")
            sys.exit(1)


        umi_header = [
            "Barcode",
            "Sample",
            "Barcoded_Reads",
            "Trimmed_Reads",
            "Unmapped_Reads",
            "Mapped_Reads",
            "Uniquely_Mapped_Reads",
            "Multimapped_Reads",
            "Chimeric_Reads",
            "No_of_Reads_w_UMIs",
            "Reads_After_Dedup_nUMIs_USSs",
            "Exon_nUMIs_USSs",
            "Intron_nUMIs_USSs",
            "Gene_nUMIs_USSs",
            "Intergenic_Reads_nUMI_USSs",
            "No_of_Genes_nUMIs_USSs",
            "No_of_Transcripts_nUMIs_USSs",
            "Ribosomal_Reads",
            "Mitochondrial_Reads",
        ]

        with open("analyze_stats_5pUMI_stats.csv", "a") as f:
            print(",".join(umi_header), file=f)
            for barcode in stats_dict:
                print(stats_dict[barcode]["Barcode"], end=",", file=f)
                print(stats_dict[barcode]["Sample"], end=",", file=f)
                print(stats_dict[barcode]["Barcoded_Reads"], end=",", file=f)
                print(stats_dict[barcode]["Trimmed_Reads"], end=",", file=f)
                print(stats_dict[barcode]["Unmapped_Reads"], end=",", file=f)
                print(stats_dict[barcode]["Mapped_Reads"], end=",", file=f)
                print(stats_dict[barcode]["Uniquely_Mapped_Reads"], end=",", file=f)
                print(stats_dict[barcode]["Multimapped_Reads"], end=",", file=f)
                print(stats_dict[barcode]["Chimeric_Reads"], end=",", file=f)
                print(stats_dict[barcode]["No_of_Reads_w_UMIs"], end=",", file=f)
                print(
                    stats_dict[barcode]["Reads_After_Dedup_nUMIs_USSs"], end=",", file=f
                )
                print(stats_dict[barcode]["Exon_nUMIs_USSs"], end=",", file=f)
                print(stats_dict[barcode]["Intron_nUMIs_USSs"], end=",", file=f)
                print(stats_dict[barcode]["Gene_nUMIs_USSs"], end=",", file=f)
                print(stats_dict[barcode]["Intergenic_Reads_nUMI_USSs"], end=",", file=f)
                print(stats_dict[barcode]["No_of_Genes_nUMIs_USSs"], end=",", file=f)
                print(
                    stats_dict[barcode]["No_of_Transcripts_nUMIs_USSs"], end=",", file=f
                )
                print(stats_dict[barcode]["Ribosomal_Reads"], end=",", file=f)
                print(stats_dict[barcode]["Mitochondrial_Reads"], end="\n", file=f)

    # write out the stats file

    output_file = "analyze_stats.csv"

    with open(output_file, "a") as f:
        print(",".join(header), file=f)
        for barcode in stats_dict:
            print(stats_dict[barcode]["Barcode"], end=",", file=f)
            print(stats_dict[barcode]["Sample"], end=",", file=f)
            print(stats_dict[barcode]["Barcoded_Reads"], end=",", file=f)
            print(stats_dict[barcode]["Trimmed_Reads"], end=",", file=f)
            print(stats_dict[barcode]["Unmapped_Reads"], end=",", file=f)
            print(stats_dict[barcode]["Mapped_Reads"], end=",", file=f)
            print(stats_dict[barcode]["Uniquely_Mapped_Reads"], end=",", file=f)
            print(stats_dict[barcode]["Multimapped_Reads"], end=",", file=f)
            print(stats_dict[barcode]["Chimeric_Reads"], end=",", file=f)
            if exp_type == "stranded_umi":
                print(stats_dict[barcode]["No_of_Reads_w_UMIs"], end=",", file=f)
                print(
                    stats_dict[barcode]["Reads_After_Dedup_nUMIs_USSs"], end=",", file=f
                )
            print(stats_dict[barcode]["Exon_Reads"], end=",", file=f)
            print(stats_dict[barcode]["Intron_Reads"], end=",", file=f)
            print(stats_dict[barcode]["Gene_Reads"], end=",", file=f)
            print(stats_dict[barcode]["Intergenic_Reads"], end=",", file=f)
            print(stats_dict[barcode]["No_of_Genes"], end=",", file=f)
            print(stats_dict[barcode]["No_of_Transcripts"], end=",", file=f)
            print(stats_dict[barcode]["Ribosomal_Reads"], end=",", file=f)

            if "Strand_Specificity" in header:
                print(stats_dict[barcode]["Mitochondrial_Reads"], end=",", file=f)
                print(stats_dict[barcode]["Strand_Specificity"], end="\n", file=f)
            else:
                print(stats_dict[barcode]["Mitochondrial_Reads"], end="\n", file=f)


if __name__ == "__main__":
    main(sys.argv[1:])
