#!/usr/bin/env python3

#
# Author: Bioinformatics group, Takara Bio USA, Inc.
#
# Script to generate reformat gene/transcript matrix file created by salmon
# This script receives the output file from salmon quantmerge as input
#
# -----------------------------------------------------------------------------
import os
import re
import sys
import argparse
import pathlib
import pandas as pd
import glob
import dask.dataframe as dd

DESCRIPTION = """Create gene/transcript matrix file output from salmon results."""

###
# Script information
###

# script info
SCRIPT_NAME = "create_genematrix.py"
VERSION = "3.1.22"

###
# Parse and define arguments
###

def transpose_csv_out_of_core(csv_path, output_csv_path, delimiter=','):
    """
    :param csv_path: 
    :param output_csv_path: 
    :param delimiter: 
    :return: 
    """
    import csv

    transposed_iterator = zip(*csv.reader(open(csv_path)))
    with open(output_csv_path, 'w') as out:
        for row in transposed_iterator:
            out.write(delimiter.join(row) + '\n')

def _parse_command_line_args(sys_argv):
    """Takes in command line parameters and parses them for use

    Args:
        sys_argv: command line input

    Returns:
        args (list) : input argument values
    """
    parser = argparse.ArgumentParser(
        description=DESCRIPTION, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    required = parser.add_argument_group("required arguments")

    required.add_argument(
        "-i",
        "--input_dir",
        dest="input_dir",
        help="Full path to output file from salmon quantmerge",
        required=True,
    )

    required.add_argument(
        "-t",
        "--type",
        dest="type",
        help="File type - one of genes/transcripts/introns.",
        required=True,
    )

    required.add_argument(
        "-o",
        "--output_file",
        dest="output_file",
        help="Name of the result file to be created",
        required=True,
    )

    args, unknown = parser.parse_known_args(sys_argv)
    return args


def main(sys_argv):
    args = _parse_command_line_args(sys_argv)

    input_dir = args.input_dir
    type = args.type
    outfile_name = args.output_file
    #neg_index = 0
    #col_replace_text = "_trimmed"

    # correct for umi-specific text in folder names
    #if "umis" in outfile_name:
    #    neg_index = 1
    #    col_replace_text = "_trimmed_umionly"

    #header_start = ""

    #header_start = "TranscriptID"
    result_file = "quant.sf"

    salmon_result_files = os.path.join(input_dir, "*_" + result_file)

    #salmon_results_list = glob.glob(salmon_result_files, recursive=True)

    #sample_id = ''
    #i = 0
    #fields = []
    #df_list = []

    # # read the output files in
    # for current_file in salmon_results_list:
    #     #current_file = os.path.join(input_dir, folder, result_file)
    #     sample_name = current_file.split("/")[-1].replace("_" + result_file, "")
    #     if re.match(r'(.*)[ACGTN]{16,24}(.*)', sample_name):
    #         if re.match(r'.*_L00\d(_trimmed)?$', sample_name):
    #             sample_id = sample_name.split("_")[-3 - neg_index]
    #         else:
    #             sample_id = sample_name.split("_")[-2 - neg_index]
    #     else:
    #         sample_id = sample_name.replace(col_replace_text, "")
    #     fields = ['TranscriptID', 'Length', 'Effective_Length', 'TPM', sample_id]
    #     # save the gene names from the first csv file
    #     if i == 0:
    #         temp_df = dd.read_csv(current_file, header=None, 
    #                               skiprows=[0,], names=fields, 
    #                               usecols=['TranscriptID',sample_id], 
    #                               dtype={'TranscriptID': 'string', 
    #                                      sample_id: 'string'}, 
    #                               sep='\t', 
    #                               assume_missing=True,
    #                               low_memory=False,
    #                               blocksize='1MB')            
    #         df_list.append(temp_df)
    #         i = 1
    #     else:
    #         temp_df = dd.read_csv(current_file, header=None, skiprows=[0,], 
    #                               names=fields, usecols=[sample_id], 
    #                               dtype={sample_id: 'string'},
    #                               sep='\t', 
    #                               assume_missing=True,
    #                               low_memory=False,
    #                               blocksize='1MB')
    #         df_list.append(temp_df)

    # sort column order by barcode name
    #final_df = dd.concat(df_list, axis=1)
    
    quants_df = dd.read_csv(salmon_result_files, sep=',', low_memory=False,
                            header=None)
    
    header_file_path = os.path.join(input_dir, "*_header.sf")
    header_results_list = glob.glob(header_file_path, recursive=True)
    header_file = header_results_list[0]
    header_df = dd.read_csv(header_file, sep=',',
                            header=None)
    
    df_list = [header_df, quants_df]
    merged_df = dd.concat(df_list, axis=0)
    
    #final_array = merged_df.to_dask_array(lengths=True)
    #transposed_array = final_array.T
    #final_df = dd.from_array(transposed_array, columns=[f'col_{i}' for i in range(transposed_array.shape[0])])

    

    # write the output file
    merged_df.to_csv(
        "temp_tobetransposed.csv", single_file=True, compute=True, encoding="utf-8", index=False
    )
    
    transpose_csv_out_of_core('temp_tobetransposed.csv', outfile_name)


if __name__ == "__main__":
    
    from dask.distributed import Client

    client = Client(n_workers=8, threads_per_worker=2, memory_limit="48GB")
    client
    main(sys.argv[1:])
