###### Loading libraries
library(Matrix)
library(Seurat)
library(data.table)
library(dplyr)
#####

##### read user arguments 
args <- commandArgs(trailingOnly = TRUE)
output_dir = args[1]
gene_transcript_map_file = args[2]
mito_genes = args[3]
reads_threshold = as.numeric(args[4])
#########################

#### read Seurat Objects
sc_exon_plus_intron_file = paste0(output_dir, "/", "CogentAP_seuratObject_gene_Exon_plus_Intron.rds")
sc_exon_plus_intron_obj = readRDS(sc_exon_plus_intron_file)

sc_exon_only_file = paste0(output_dir, "/", "CogentAP_seuratObject_gene_Exon_Only.rds")
sc_exon_only_obj = readRDS(sc_exon_only_file)

sc_isforms_file = paste0(output_dir, "/", "CogentAP_seuratObject_isoforms_Only.rds")
sc_isoforms_obj = readRDS(sc_isforms_file)
################

####### Calculate Number of genes / isoforms detected
mito_genes_list = read.table(mito_genes, header = FALSE, sep = "\n", check.names = FALSE)
mito_genes_list = mito_genes_list$V1

gene_transcript_map = read.table(gene_transcript_map_file, header = TRUE, sep = "\t", row.names = 1, check.names = FALSE)
gene_transcript_map_mito = gene_transcript_map %>% dplyr::filter(Gene_ID %in% mito_genes_list)
mito_genes = unique(gene_transcript_map_mito$Gene_Name)

sc_exon_plus_intron_obj_mat = sc_exon_plus_intron_obj[['RNA']]$counts
sc_exon_plus_intron_obj_mat_mito = sc_exon_plus_intron_obj_mat[mito_genes,]

sc_isoforms_obj_mat = sc_isoforms_obj[['RNA']]$counts
sc_isoforms_obj_mat = sc_isoforms_obj_mat[,colnames(sc_exon_plus_intron_obj_mat)]

diff_thres = reads_threshold - 0

genes_detected = c()
mito_detected = c()
isoform_detected = c()
for(f in 1:diff_thres) {
	value = f-1
	colname_use = paste0("counts_", value)
	temp_genes = data.frame(sparseMatrixStats::colCounts(sc_exon_plus_intron_obj_mat, value = (f-1)))
	colnames(temp_genes) = colname_use
	temp_mito = data.frame(sparseMatrixStats::colCounts(sc_exon_plus_intron_obj_mat_mito, value = (f-1)))
	colnames(temp_mito) = colname_use
	temp_isoform = data.frame(sparseMatrixStats::colCounts(sc_isoforms_obj_mat, value = (f-1)))
	colnames(temp_isoform) = colname_use
	genes_detected = c(genes_detected, temp_genes)
	mito_detected = c(mito_detected, temp_mito)
	isoform_detected = c(isoform_detected, temp_isoform)
}
final_genes_detected = do.call(cbind, genes_detected)
final_genes_detected = data.frame(final_genes_detected)
final_genes_detected$sum = rowSums(final_genes_detected)
final_genes_detected$num_genes_detected = nrow(sc_exon_plus_intron_obj_mat) - final_genes_detected$sum
rownames(final_genes_detected) = colnames(sc_exon_plus_intron_obj_mat)
final_mito_detected = do.call(cbind, mito_detected)
final_mito_detected = data.frame(final_mito_detected)
final_mito_detected$sum = rowSums(final_mito_detected)
final_mito_detected$num_mito_detected = nrow(sc_exon_plus_intron_obj_mat_mito) - final_mito_detected$sum
rownames(final_mito_detected) = colnames(sc_exon_plus_intron_obj_mat)
final_isoform_detected = do.call(cbind, isoform_detected)
final_isoform_detected = data.frame(final_isoform_detected)
final_isoform_detected$sum = rowSums(final_isoform_detected)
final_isoform_detected$num_isoform_detected = nrow(sc_isoforms_obj_mat) - final_isoform_detected$sum
rownames(final_isoform_detected) = colnames(sc_exon_plus_intron_obj_mat)
##############################


##### Calculate Barcode Stats

df_exon_plus_intron = data.frame(colSums(sc_exon_plus_intron_obj[['RNA']]$counts))
colnames(df_exon_plus_intron) = "Exon_plus_Intron_Counts"

df_mito_reads = data.frame(colSums(sc_exon_plus_intron_obj_mat_mito[,colnames(sc_exon_plus_intron_obj[['RNA']]$counts)]))
colnames(df_mito_reads) = "Mitochondrial_Reads"

df_exon = data.frame(colSums(sc_exon_only_obj[['RNA']]$counts[,colnames(sc_exon_plus_intron_obj[['RNA']]$counts)]))
colnames(df_exon) = "Exon_Only_Counts"

final_df_stats = cbind(df_exon_plus_intron, df_exon, df_mito_reads)
final_df_stats$Intron_Only_Counts = final_df_stats$Exon_plus_Intron_Counts - final_df_stats$Exon_Only_Counts

##### Add stats for gene, mito genes and isoform detection
final_genes_detected = final_genes_detected[rownames(final_df_stats),]
final_mito_detected = final_mito_detected[rownames(final_df_stats),]
final_isoform_detected = final_isoform_detected[rownames(final_isoform_detected),]
final_df_stats$Number_of_Genes_Expressed = final_genes_detected$num_genes_detected
final_df_stats$Number_of_MitoGenes_Expressed = final_mito_detected$num_mito_detected
final_df_stats$Number_of_Isoforms_Expressed = final_isoform_detected$num_isoform_detected
##### write the stats
stats_out_file = paste0(output_dir, "/", 'barcode_stats.csv')
write.csv(final_df_stats, stats_out_file, quote = F)
##################

####### Generate a seurat object with following: RNA Assay (Exon+Intron Counts); Exon_RNA (Exon Only Counts); and Transcript_RNA (Isoform Counts)
sc_exon_plus_intron_obj[['Exon_RNA']] = CreateAssay5Object(sc_exon_only_obj[['RNA']]$counts)
sc_exon_plus_intron_obj[['Transcript_RNA']] = CreateAssay5Object(sc_isoforms_obj[['RNA']]$counts)

outfile = paste0(output_dir, "/", 'CogentAP_gene_and_isoform_counts.rds')
saveRDS(sc_exon_plus_intron_obj, file = outfile)
#################

#### Clean up by moving individual objects to new folder!
individual_data_storage = paste0(output_dir, "/", "seurat_objs_per_count_files")
dir.create(individual_data_storage)
file.copy(sc_exon_plus_intron_file, individual_data_storage)
file.copy(sc_exon_only_file, individual_data_storage)
file.copy(sc_isforms_file, individual_data_storage)
unlink(sc_exon_plus_intron_file)
unlink(sc_exon_only_file)
unlink(sc_isforms_file)
#####
