###### Loading libraries
library(Matrix)
library(Seurat)
library(data.table)
library(dplyr)
library(tidyverse)
#####

##### read user arguments 
args <- commandArgs(trailingOnly = TRUE)
dir_salmon = args[1]
gene_transcript_map = args[2]
gene_info_file = args[3]
transcript_info_file = args[4]
output_dir = args[5]
#########################

##### reading gene info & transript info
gene_info = read.table(gene_info_file, header = T, row.names = 1, check.names = F, sep = "\t") ## read gene info file for later use
transcript_info = read.csv(transcript_info_file, header = T, row.names = 1, check.names = F) ## read gene info file for later use
transcript_info$ID = rownames(transcript_info)
######

###### create a tmp directory
tmp_dir = paste0(output_dir, "/", "tmp_geneExon")
dir.create(tmp_dir)
######

###### read quant files and their corresponding paths
message("Reading file names along with full path")
file_paths <- list.files(path = dir_salmon, pattern = "_quant.sf$", recursive = TRUE, full.names = TRUE)
#######

####### read header file for transcript info
file_paths_header <- list.files(path = dir_salmon, pattern = "_header.sf$", recursive = TRUE, full.names = TRUE)
transcripts_read = read.table(file_paths_header[1], header = TRUE, sep = "\n", check.names = FALSE)
transcripts_header = transcripts_read$TranscriptID
#########

####### read gtf transcript - gene map
gtf_trans_select = read.table(gene_transcript_map, sep = "\t", header = T, row.names = 1, check.names = F)
gtf_trans_select = gtf_trans_select[transcripts_header,]
transcripts = unique(gtf_trans_select$transcript_id)
transcripts_intronicExon = grep('Intronic_Transcript', transcripts, value = T)
transcripts_exonOnly = transcripts[! transcripts %in% transcripts_intronicExon]
gtf_trans_select_exonOnly = gtf_trans_select[transcripts_exonOnly,]
######################################

###### create 100 chunks based on total quant files
total_cols = length(file_paths) ## find total numbe rof columns in data

cells_list <- 1:total_cols ## create a cell-list 

chunk_size <- round(total_cols/100) # Define the number of elements you want in each chunk (10)

if(chunk_size == 0) {
	chunks <- split(cells_list, 1)
} else {
	breakpts <- seq(0, length(cells_list)+(length(cells_list) %% 10), by = chunk_size) # Generate breakpoints for cutting the vector into chunks starting from 0 

	chunks <- cut(seq_along(cells_list), breaks = breakpts, labels = FALSE) # Cut the vector into chunks based on the breakpoints 

	chunks <- split(cells_list, chunks) # Split the vector into chunks based on the cuts

	if(chunks[[length(chunks)]][length(chunks[[length(chunks)]])] == total_cols) {
		chunks = chunks
	}
	if(chunks[[length(chunks)]][length(chunks[[length(chunks)]])] < total_cols) {
		use_set = c(chunks[[length(chunks)]][1]:(total_cols))
		chunks[[length(chunks)]] = use_set
	}
}
##########

###### process each chunk 
sparse_mat_gene_exonOnly = c()
sparse_mat_transcripts = c()
gene_ids_store_exon = c()
transcript_ids_store = c()

str = paste0('Number of chucks to process: ', length(chunks))
message(str)
for(f in 1:length(chunks)) {

	str = paste0('Executing chunk - ', f)
	message(str)

	#### set file list in each chunk
	start_cell = chunks[[f]][1]
	length_chunk = length(chunks[[f]])
	end_cell = chunks[[f]][length_chunk]
	file_list = file_paths[start_cell:end_cell]

	str = paste0('Number of cell files to process in chunk: ', f, ' is: ', length_chunk)
	message(str)
	
	#### setup a execution to write a matrix using file list
	list = paste(file_list, collapse = " ")
	temp_file = paste0(tmp_dir, "/", "chunk-", f, ".txt")
	cmd_execute = paste('paste', list, '>', temp_file, collapse = "")
	system(cmd_execute)
	message('wrote chunk associated matrix in tmp folder')

	#### read the matrix written and setting up data frames
	trans_counts = fread(temp_file, data.table = FALSE, sep = "\t", check.names = FALSE)
	rownames(trans_counts) = transcripts
	message('read chunk associated matrix from tmp folder')
	
	if(length(colnames(trans_counts)) == 1) {
		trans_counts_exonOnly = data.frame(trans_counts[transcripts_exonOnly,])
		rownames(trans_counts_exonOnly) = transcripts_exonOnly
		colnames(trans_counts_exonOnly) = colnames(trans_counts)
		trans_counts_exonOnly_copy = trans_counts_exonOnly
		trans_counts_exonOnly$Gene_Name = gtf_trans_select_exonOnly$Gene_Name
	} else {
		trans_counts_exonOnly = trans_counts[transcripts_exonOnly,]
		trans_counts_exonOnly_copy = trans_counts_exonOnly
		trans_counts_exonOnly$Gene_Name = gtf_trans_select_exonOnly$Gene_Name
	}

	message('Summarizing gene counts for exons Only.. please wait.')
	geneCounts_exonOnly_df <- trans_counts_exonOnly %>% group_by(Gene_Name) %>% summarise(across(where(is.numeric), sum)) %>% as.data.frame()

	gene_ids_store_exon = geneCounts_exonOnly_df[,1]
	geneCounts_exonOnly_df[,1] = NULL
	transcript_ids_store = rownames(trans_counts_exonOnly_copy)
	rownames(trans_counts_exonOnly_copy) = NULL

	geneCounts_exonOnly_mat = as.matrix(geneCounts_exonOnly_df)
	trans_counts_exonOnly_mat = as.matrix(trans_counts_exonOnly_copy)

	geneCounts_exonOnly_mat = Matrix::Matrix(geneCounts_exonOnly_mat, sparse = T)
	trans_counts_exonOnly_mat = Matrix::Matrix(trans_counts_exonOnly_mat, sparse = T)

	sparse_mat_gene_exonOnly = c(sparse_mat_gene_exonOnly, geneCounts_exonOnly_mat)
	sparse_mat_transcripts = c(sparse_mat_transcripts, trans_counts_exonOnly_mat)
	str = paste0('Execution Complete for chunk - ', f)
	message(str)
}

######## bind sparse matrices and generating barcode level counts info
message('finalizing sparse matrices')

final_sparse_mat_genes_exon_only = do.call(cbind, sparse_mat_gene_exonOnly) ## cbind all sparse matrices
final_sparse_mat_transcripts = do.call(cbind, sparse_mat_transcripts) ## cbind all sparse matrices

rownames(final_sparse_mat_genes_exon_only) = gene_ids_store_exon
rownames(final_sparse_mat_transcripts) = transcript_ids_store

barcode_ids = colnames(final_sparse_mat_genes_exon_only)

check_barcode_presence = grep('[ACGTN]{16,24}', barcode_ids, value = TRUE)

if(length(check_barcode_presence) == length(barcode_ids)) {
	message('detected single cell chemistry...finding barcodes ids..')
	split_barcodes = strsplit(barcode_ids, split = "_", fixed = TRUE)
	barcodes_final = c()
	for(f in 1:length(split_barcodes)) {
		barcode_temp = split_barcodes[[f]][which(!is.na(str_extract(split_barcodes[[f]], regex('[ACGTN]{16,24}'))))]
		barcodes_final = c(barcodes_final, barcode_temp)
	}
} else {
	message('detected bulk RNA cell chemistry...finding sample ids..')
	split_barcodes = strsplit(barcode_ids, split = "_trimmed", fixed = TRUE) ### splitting at _trimmed
	barcodes_temp1 = sapply(split_barcodes, "[[", 1, simplify = TRUE)
	barcodes_final = barcodes_temp1
}

colnames(final_sparse_mat_genes_exon_only) = barcodes_final
colnames(final_sparse_mat_transcripts) = barcodes_final
###########

######### Initializing seurat object generation
message('Initializing seurat object generation')

seurat_object_exon = CreateSeuratObject(final_sparse_mat_genes_exon_only)
###############################################

########## adding transcript names to Isoform Matrix & writing Seurat Object
FOLDER_NAME = paste0(tmp_dir, '/', 'Isoform_LARGE_CSV_MM/')
dir.create(FOLDER_NAME) ## create a drirectory for resembling 10X out

mtx_file = paste0(FOLDER_NAME, "matrix.mtx")
gene_file = paste0(FOLDER_NAME, "genes.tsv")
barcodes_files = paste0(FOLDER_NAME, "barcodes.tsv")

writeMM(obj = final_sparse_mat_transcripts, file = mtx_file) ## writing a Matrix object
write(x = rownames(final_sparse_mat_transcripts), file = gene_file) ## writing genes.tsv
write(x = colnames(final_sparse_mat_transcripts), file = barcodes_files) ## writing barcodes.tsv
gene_tsv = read.table(gene_file, sep = "\t", header = F) ## re-read genes.tsv
colnames(gene_tsv) = c("names") ## add column name
gene_info_ordered = c()
for(f in 1:length(colnames(transcript_info)))
{
	check_value = match(gene_tsv$names, transcript_info[,f])
	if(any(is.na(check_value)) == TRUE)
	{
		print ("checking..")
	}
	else
	{
		print ("Match found, ordering dataframe")
		gene_info_ordered = transcript_info[match(gene_tsv$names, transcript_info[,f]),]
	}
}
gene_tsv$gene_name = gene_info_ordered$Transcript_Name ## add gene names for corresponding ens ids
gene_tsv_new = cbind(gene_tsv$gene_name, gene_tsv$gene_name) ## create a new data.frame with two gene name cols
write.table(gene_tsv_new, gene_file, sep = "\t", quote = F, col.names = F, row.names = F) ## write the new genes.tsv
isoform_counts = Read10X(FOLDER_NAME)
seurat_object_isoform = CreateSeuratObject(isoform_counts)
###############################################

##### Save seurat object with following Counts:
outfile = paste0(output_dir, "/", "CogentAP_seuratObject_gene_Exon_Only.rds")
saveRDS(seurat_object_exon, file = outfile)

outfile = paste0(output_dir, "/", "CogentAP_seuratObject_isoforms_Only.rds")
saveRDS(seurat_object_isoform, file = outfile)
######################################################


##### starting clean up!
unlink(tmp_dir, recursive = TRUE)
message('All done!')
#######
