
############# Check and Load Required Packages ###################
message('Checking and loading necessary packages...')
required_packages <- c("Seurat", "dplyr", "data.table", "optparse")
for (pkg in required_packages) {
  suppressPackageStartupMessages(library(pkg, character.only = TRUE))
}
##################################################################

############### Command-Line Argument Parsing ####################
message('Parsing command-line arguments...')
option_list <- list(
  make_option("--pathtoCogentDSreporter", type="character", default=NULL, help="Path to CogentDS reporter script", metavar="character"),
  make_option("--gc_file", type = "character", default = NULL, help = "Path to GC content file", metavar = "character"),
  make_option("--ginkgo_file", type = "character", default = NULL, help = "Path to ginkgo data file", metavar = "character"),
  make_option("--ccn_file", type = "character", default = NULL, help = "Path to SegCopy file", metavar = "character"),
  make_option("--bounds_file", type = "character", default = NULL, help = "Path to bounds file", metavar = "character"),
  make_option("--outPath", type="character", default=NULL, help="Path to write Output files", metavar="character"),  
  make_option("--outCCNObject", type="character", default='CogentDS_analysis.rds', help="Output Seurat v5 Object", metavar="character"),
  make_option("--outHTMLReport", type="character", default='CogentDS_preliminary-analysis_report.html', help="HTML report of analysis", metavar="character")  
)

opt_parser <- OptionParser(option_list = option_list)
opt <- parse_args(opt_parser)

# Check for required arguments
if (is.null(opt$gc_file) || is.null(opt$ginkgo_file) || is.null(opt$ccn_file) || is.null(opt$bounds_file)) {
  print_help(opt_parser)
  stop("ERROR: All input file paths (gc_file, ginkgo_file, ccn_file, bounds_file) must be provided.", call. = FALSE)
}

########### Create path to required scripts ##########
if(!is.null(opt$pathtoCogentDSreporter)) {
  Generate_HTML_report = paste0(opt$pathtoCogentDSreporter, "/bin/Generate_HTML_report.R")
  if(!is.null(opt$outPath)) {
    dir.create(opt$outPath)
    temppath = paste0(opt$outPath, "/temp/")
    dir.create(temppath)
  } else {
    stop("--outPath is set to NULL!! value needed for writing files is specified directory")
  }
} else {
  stop("FATAL ERROR: Path to 'CogentDS_DataHandler_and_Reporter.R' is missing. Argument not supplied.. Found NULL", call.=FALSE)
}
########################################################

##################################################################

############### Load Data Files ##################################
message('Loading data files...')
GC <- read.table(opt$gc_file, header = FALSE, sep = "\t", as.is = TRUE)
colnames(GC) <- "GC_Content"

ginkgo_data <- read.table(opt$ginkgo_file, header = TRUE, sep = "\t")
ccn <- read.table(opt$ccn_file, header = TRUE)
rownames(ccn) <- paste(ccn$CHR, ccn$START, ccn$END, sep = "-")
bounds <- read.table(opt$bounds_file)
##################################################################

############### Create Seurat Object and Populate ################
message('Creating Seurat object...')
ccnObject <- Seurat::CreateSeuratObject(counts = ccn[, 4:ncol(ccn)])
ccnObject[["ginkgo_data"]] <- Seurat::CreateAssayObject(counts = ginkgo_data)

# Adding additional metadata
ccnObject[["RNA"]][["Chrom_info"]] <- paste(ccn$CHR, ccn$START, ccn$END, sep = "-")
ccnObject[["RNA"]][["GC_Content"]] <- GC
ccnObject@misc$bounds <- bounds
##################################################################

################## Save Seurat Object ############################
message('Saving Seurat object to disk...')
outfile = paste0(opt$outPath, "/", opt$outCCNObject)
saveRDS(ccnObject, file = outfile)
##################################################################

#########################################################################

################ Generate DS Report from Seurat Object ##################
message('Generating Report..')
if(!is.null(ccnObject))
{
  ccnDSobject_file = paste0(opt$outPath, "/", opt$outCCNObject)
  source(Generate_HTML_report, local = TRUE)
  file_path = paste0(opt$outPath, "/", opt$outHTMLReport)
  if (file.exists(file_path)) {
    message('Object created and report generated')
  } else {
    message('Report generation failed')
  }
  unlink(temppath, recursive = TRUE)
} else {
  stop('Object generation has failed! Please check inputs')
}
