#!/usr/bin/env nextflow

nextflow.enable.dsl = 2

// import relevant module
include { FASTQC } from '../modules/fastqc'
include { CUTADAPT } from '../modules/cutadapt'
include { FASTQC_TRIM } from '../modules/fastqc_trim'
include { SORTMERNA } from '../modules/sortmerna'
include { STAR_ALIGN } from '../modules/star_align'
include { STAR_LOAD_GENOME } from '../modules/star_load_genome'
include { DEDUP_UMITOOLS;
    DEDUP_UMITOOLS as DEDUP_UMITOOLS_UMIONLY } from '../modules/dedup_umitools'
include { SALMON_QUANT;
    SALMON_QUANT as SALMON_QUANT_UMIONLY } from '../modules/salmon_quant'
include  { PROCESS_EXON_INTRON; 
    PROCESS_EXON_INTRON as PROCESS_EXON_INTRON_UMIONLY } from '../modules/process_exon_intron'
include { PROCESS_EXON_ISOFORMS; 
    PROCESS_EXON_ISOFORMS as PROCESS_EXON_ISOFORMS_UMIONLY } from '../modules/process_exon_isoforms'
include { FINALIZE_SEURAT_OBJECTS;
    FINALIZE_SEURAT_OBJECTS as FINALIZE_SEURAT_OBJECTS_UMIONLY } from '../modules/finalize_seurat_objects'
include { FUSION_WORKFLOW } from './fusion_workflow'
include { IMMUNE_WORKFLOW } from './immune_workflow'
include { CREATE_ANALYZE_STATS } from '../modules/create_analyze_stats'
include { COGENTDS_REPORT;
    COGENTDS_REPORT as COGENTDS_REPORT_UMIONLY } from '../modules/cogentds_report'

workflow RNA_ANALYZE_WORKFLOW {

    take:
        input_fastq_ch

    main:

        transcript_fa = file("${params.genome_dir}/${params.genome}/${params.genome}.intron.transcripts.fa")
        transcript_info_file = file("${params.genome_dir}/${params.genome}/transcript_info.csv")
        gene_biotype_file = file("${params.genome_dir}/${params.genome}/gene_biotype_and_length_info.txt")
        gene_map_file = file("${params.genome_dir}/${params.genome}/gene_transcript_map_with_genename.txt")
        gene_info_file = file("${params.genome_dir}/${params.genome}/gene_info.csv")
        star_index = file("${params.genome_dir}/${params.genome}/star_index")


        // only validate genome paths if it is not in a s3 bucket
        if ( !params.genome_dir.startsWith('s3:') ) {
            if ( !transcript_fa.exists() ) {
                exit 1, "Transcriptome file does not exist, please run add_genome first"
            }
            if ( !transcript_info_file.exists() ) {
                exit 1, "Transcript info file does not exist, please run add_genome first"
            }
            if ( !gene_map_file.exists() ) {
                exit 1, "Gene map file does not exist, please run add_genome first"
            }
            if ( !gene_biotype_file.exists() ) {
                exit 1, "Gene biotype file does not exist, please run add_genome first"
            }
            if ( !gene_info_file.exists() ) {
                exit 1, "Gene info file does not exist, please run add_genome first"
            }
            if ( !star_index.exists() ) {
                exit 1, "STAR Index path does not exist, please run add_genome first"
            }
        }

        Channel.fromPath(transcript_fa)
            .ifEmpty { error "Transcriptome fasta file does not exist, please run add_genome first" }
            .set{ transcript_fa_ch }

        Channel.fromPath(star_index)
            .ifEmpty { error "STAR Index path does not exist, please run add_genome first" }
            .set{ star_index_ch }

        Channel.fromPath(gene_map_file)
            .ifEmpty { error "Gene map file does not exist, please run add_genome first" }
            .set{ gene_map_file_ch }

        Channel.fromPath(transcript_info_file)
            .ifEmpty { error "Transcript info file does not exist, please run add_genome first" }
            .set{ transcript_info_file_ch }

        Channel.fromPath(gene_info_file)
            .ifEmpty { error "Gene info file does not exist, please run add_genome first" }
            .set{ gene_info_file_ch }

        Channel.fromPath(gene_biotype_file)
            .ifEmpty { error "Gene Biotype file does not exist, please run add_genome first" }
            .set{ gene_biotype_file_ch }

        CUTADAPT(input_fastq_ch)

        cutadapt_log_ch = Channel.empty()
        CUTADAPT.out.json
            .collect()
            .set{ cutadapt_log_ch }

        STAR_LOAD_GENOME(star_index_ch, cutadapt_log_ch)

        def fastq_trimmed = CUTADAPT.out.trimmed_reads

        fastq_trimmed
            .flatten()
            .map{ file ->
                    def sample = file.name.replaceAll("_R[12].fastq.gz", "")
                    tuple(sample, file)
            }
            .groupTuple()
            .set{ fastq_trimmed_ch }

        if ( params.fastqc ) {
            FASTQC(input_fastq_ch)
            FASTQC_TRIM(fastq_trimmed_ch)
        }

        // check if rRNA removal is needed

        sortmerna_log_ch = Channel.empty()
        clonotype_matrix_ch = Channel.empty()
        clonotype_summary_ch = Channel.empty()
        clonotype_metadata_ch = Channel.empty()

        if ( params.type_of_experiment == "shasta_total_rna" && params.ribodepletion != true ) {

            fastq_trimmed_ch
                .combine(star_index_ch)
                .set{ star_align_input_ch }

            if ( params.immune ){

                IMMUNE_WORKFLOW(fastq_trimmed_ch)
            }

        } else if ( params.ribodepletion == false) {
            
            fastq_trimmed_ch
                .combine(star_index_ch)
                .set{ star_align_input_ch }

            if ( params.immune ){

                IMMUNE_WORKFLOW(fastq_trimmed_ch)
            }
        } else {

            ribo_db = file("${params.genome_dir}/${params.genome}/sortmernadb/idx", checkIfExists: true)

            sortmerna_index_ch = Channel.fromPath(ribo_db)

            Channel.fromPath("${params.genome_dir}/${params.genome}/sortmernadb/ref/*")
                .ifEmpty { error "Could not find ribosomal database fasta files!" }
                .collect()
                .set { sortmerna_fastas_ch }

            fastq_trimmed_ch
                .combine(sortmerna_index_ch)
                .set{ sortmerna_input_ch }

            SORTMERNA(sortmerna_input_ch, sortmerna_fastas_ch)

            SORTMERNA.out.ribodepleted_fastqs
                .combine(star_index_ch)
                .set{ star_align_input_ch }

            SORTMERNA.out.log
                .collect()
                .set{ sortmerna_log_ch }


            if ( params.immune ){

                IMMUNE_WORKFLOW(SORTMERNA.out.ribodepleted_fastqs)

            }

        }

        if ( params.immune ){
            clonotype_matrix_ch = IMMUNE_WORKFLOW.out.clonotype_matrix
            clonotype_summary_ch = IMMUNE_WORKFLOW.out.immune_summary
            clonotype_metadata_ch = IMMUNE_WORKFLOW.out.immune_metadata
        }

        star_log_ch = Channel.empty()
        star_reads_ch = Channel.empty()
        umitools_log_ch = Channel.empty()
        salmon_log_ch = Channel.empty()
        final_counts_rds_ch = Channel.empty()
        barcode_stats_ch = Channel.empty()
        fusion_results_ch = Channel.empty()

        def umi_only = false

        STAR_ALIGN(star_align_input_ch)

        if (params.fusion){
                FUSION_WORKFLOW(STAR_ALIGN.out.junction_file)

                fusion_results_ch = FUSION_WORKFLOW.out.fusion_results
        }

        STAR_ALIGN.out.log_final_file
            .collect()
            .set{ star_log_ch }

        STAR_ALIGN.out.read_counts_file
            .collect()
            .set{ star_reads_ch }

        if (params.type_of_experiment == 'stranded_umi'){

            DEDUP_UMITOOLS(STAR_ALIGN.out.transcript_bam)

            DEDUP_UMITOOLS.out.dedup_bam
                .combine(transcript_fa_ch)
                .set { salmon_input_ch }

            DEDUP_UMITOOLS.out.umitools_log
                .collect()
                .set{ umitools_log_ch }

        } else {
            STAR_ALIGN.out.transcript_bam
                .combine(transcript_fa_ch)
                .set { salmon_input_ch }

        }

        SALMON_QUANT(salmon_input_ch, umi_only)

        SALMON_QUANT.out.salmon_transcript_results
            .collect()
            .set{ salmon_transcript_results_ch }

        SALMON_QUANT.out.salmon_logs
            .collect()
            .set{ salmon_log_ch }

        SALMON_QUANT.out.salmon_transcript_header
            .collect()
            .set { salmon_transcript_header_ch }

        mito_file_ch = Channel.empty()

        Channel.fromPath("${params.genome_dir}/${params.genome}/mito_reads/*.txt")
                .ifEmpty { error "Could not find mito read files!" }
                .collect()
                .set { mito_file_ch }

        PROCESS_EXON_INTRON(
            salmon_transcript_results_ch,
            salmon_transcript_header_ch,
            gene_map_file_ch,
            gene_biotype_file_ch,
            transcript_info_file_ch,
            gene_info_file_ch
        )

        PROCESS_EXON_ISOFORMS(
            salmon_transcript_results_ch,
            salmon_transcript_header_ch,
            gene_map_file_ch,
            gene_biotype_file_ch,
            transcript_info_file_ch
        )

        FINALIZE_SEURAT_OBJECTS(
            PROCESS_EXON_INTRON.out.exon_plus_intron_rds,
            PROCESS_EXON_ISOFORMS.out.isoform_rds,
            PROCESS_EXON_ISOFORMS.out.exon_rds,
            mito_file_ch.ifEmpty([]),
            gene_map_file_ch,
            umi_only
        )

        final_counts_rds_ch = FINALIZE_SEURAT_OBJECTS.out.final_counts_rds
        barcode_stats_ch = FINALIZE_SEURAT_OBJECTS.out.barcode_stats_file

        umitools_umionly_log_ch = Channel.empty()
        final_counts_umi_rds_ch = Channel.empty()
        barcode_stats_umi_ch = Channel.empty()


        if ( params.type_of_experiment == 'smartseq_fla_umi'){
            umi_only = true

            DEDUP_UMITOOLS_UMIONLY(STAR_ALIGN.out.transcript_bam)

            DEDUP_UMITOOLS_UMIONLY.out.dedup_bam
                .combine(transcript_fa_ch)
                .set { salmon_umionly_input_ch }

            DEDUP_UMITOOLS_UMIONLY.out.umitools_log
                .collect()
                .set{ umitools_umionly_log_ch }

            SALMON_QUANT_UMIONLY(salmon_umionly_input_ch, umi_only)

            SALMON_QUANT_UMIONLY.out.salmon_transcript_results
                .collect()
                .set{ salmon_umionly_transcript_results_ch }

            SALMON_QUANT_UMIONLY.out.salmon_logs
                .collect()
                .set{ salmon_umionly_log_ch }
            
            SALMON_QUANT_UMIONLY.out.salmon_transcript_header
                .collect()
                .set { salmon_umionly_transcript_header_ch }

            PROCESS_EXON_INTRON_UMIONLY(
                salmon_umionly_transcript_results_ch,
                salmon_umionly_transcript_header_ch,
                gene_map_file_ch,
                gene_biotype_file_ch,
                transcript_info_file_ch,
                gene_info_file_ch
            )

            PROCESS_EXON_ISOFORMS_UMIONLY(
                salmon_umionly_transcript_results_ch,
                salmon_umionly_transcript_header_ch,
                gene_map_file_ch,
                gene_biotype_file_ch,
                transcript_info_file_ch
            )

            FINALIZE_SEURAT_OBJECTS_UMIONLY(
                PROCESS_EXON_INTRON_UMIONLY.out.exon_plus_intron_rds,
                PROCESS_EXON_ISOFORMS_UMIONLY.out.isoform_rds,
                PROCESS_EXON_ISOFORMS_UMIONLY.out.exon_rds,
                mito_file_ch.ifEmpty([]),
                gene_map_file_ch,
                umi_only
            )
            
            final_counts_umi_rds_ch = FINALIZE_SEURAT_OBJECTS_UMIONLY.out.final_counts_rds
            barcode_stats_umi_ch = FINALIZE_SEURAT_OBJECTS_UMIONLY.out.barcode_stats_file

        }


        CREATE_ANALYZE_STATS(
            cutadapt_log_ch.ifEmpty([]),
            sortmerna_log_ch.ifEmpty([]),
            salmon_log_ch.ifEmpty([]),
            star_log_ch.ifEmpty([]),
            star_reads_ch.ifEmpty([]),
            umitools_log_ch.ifEmpty([]),
            umitools_umionly_log_ch.ifEmpty([]),
            barcode_stats_ch.ifEmpty([]),
            barcode_stats_umi_ch.ifEmpty([])
        )

        if ( params.type_of_experiment == 'smartseq_fla_umi'){

            umi_only = true

            COGENTDS_REPORT_UMIONLY(
                CREATE_ANALYZE_STATS.out.stats_file,
                final_counts_umi_rds_ch, 
                clonotype_matrix_ch.ifEmpty([]),
                clonotype_summary_ch.ifEmpty([]),
                clonotype_metadata_ch.ifEmpty([]),
                fusion_results_ch.ifEmpty([]),
                umi_only
            )
        }
        
        umi_only = false

        COGENTDS_REPORT(
            CREATE_ANALYZE_STATS.out.stats_file,
            final_counts_rds_ch, 
            clonotype_matrix_ch.ifEmpty([]),
            clonotype_summary_ch.ifEmpty([]),
            clonotype_metadata_ch.ifEmpty([]),
            fusion_results_ch.ifEmpty([]),
            umi_only
        )
}
